# Binary search and replace (REGEX) program.
# SYNTAX: BREGX <pattern> <replacement> <sourcefile> <targetfile> [switches]
# The pattern and replacement variables are internally limited to 128
# efective characters, though they need not be the same length.  The pattern
# is tested against a window in the target file that is 4096 bytes long, and
# has a 128 byte overlap with the previous and following windows, unless the
# source file is less than 4K, in which case the file is processed in its
# entirety.

# If the find only switch is used, the show addresses switch is forced also.

# command tail argument: pattern    -   The regular expression to match
# command tail argument: replacement-   The string to replace the pattern
# command tail argument: source     -   The name of the source file
# command tail argument: target     -   The name of the file to write to.
# command tail argument: offset     -   Optional number of bytes to copy
#                                       without processing. (Hex or decimal)
# command tail argument: aflag      -   /a causes print to stdout of addresses
#                                       of matches
# command tail argument: fflag      -   /f prevents substitutions (find only)
# command tail argument: count      -   /cnnn maximum number of replacements

# Return ERRORLEVELS
#   0   -   normal termination
#   1   -   no arguments
#   2   -   unrecognized switch
#   3   -   user aborted at "Replace target?" prompt
#   4   -   excessive number of arguments
#   5   -   replace mode chosen, but either replacement string or target file
#           argument (or both) missing
#   6   -   source file or pattern argument missing (or both)
#   7   -   source file does not exist or is unreadable
#   8   -   source file is empty
#   9   -   target file cannot be opened for write
#  10   -   defective /cnn argument
#  11   -   the pattern matches the replacement string

global ptrn = "", rplcmt = "", src = "", trgt = ""
global inbuf, ovrlp
global fullbuffer = 4096, lapsize = 128
global lead, endbs, enddbs, chpattern
global aflag = "no", fflag = "replace", maxc = -1
global oldlist, subs = 0, matches = 0


BEGIN{
local i = 0
local runtime, rt

#############################################################################
#                                                                           #
# Initialization code                                                       #
#                                                                           #
#############################################################################

runtime = time()

# Display banner to STDERR
    banner()

    if(ARGV[1] == "")
    {
        syntaxerr(1)     # Does not return
    }

# Use raw read and writes for binary data
    RAWMODE = 7

# Put arguments into local variables

    while(ARGV[i++] != "")  # For each non-null argument.
    {   
        scanarg(ARGV[i])

    }
# Now, see if we have what is needed, in the way of arguments
    checkargs()

# Test for existence and readability of source file
    if(fopen(src, "r") == 0)
    {
        badsource(7)     # Does not return
    }
# Get and test the length of the source file.
    sfs = filesize(src)
    if(sfs == 0)
    {
        badsource(8)     # Does not return
    }

# Test for existence of target file
    if(fflag == "replace")
    {
        if(fopen(trgt, "r") != 0)
        {
            targetxist(src, trgt)      # May not return
        }

# Display the arguments as interpreted by the above parser.  Status messages
# go to STDERR to allow redirecting the addresses without extranious matter in
# the resulting file.
    printargs()


# If we get here, the source file is open for read and the target file may be
# opened for create/truncate.

# Test for creatability of target file
        if(fopen(trgt, "w") == 0)
        {
            badtarget(9)         # Does not return
        }
    }
# At this point, we should be able to read from the source and write from the
# target files.  Both are open (source only in find mode).

#############################################################################
#                                                                           #
# Main business of the program                                              #
#                                                                           #
#############################################################################


# Compile various regular expressions.
    doregex()

# Copy the offset number of bytes without processing.
    skip(src, trgt, offset)

# Call the main function.
  dolargefile(src, trgt, ptrn, convrt(rplcmt), rplcmt)

#############################################################################
#                                                                           #
# Cleanup code                                                              #
#                                                                           #
#############################################################################

# Close files before exiting with a return code of zero.
    close(src)
    if(fflag == "replace") {close(trgt)}

# Clean up cursor position and report finish.
    cleanup(runtime)
    exit(0)
# Done!
}

